@@ -42,13 +42,14 @@ except ImportError:  # pragma: no cover (py3 only)
 
 from pkg_resources import parse_version
 
-nodeenv_version = '1.3.3'
+nodeenv_version = '1.3.4'
 
 join = os.path.join
 abspath = os.path.abspath
+iojs_taken = False
 src_domain = "nodejs.org"
 
-is_PY3 = sys.version_info[0] == 3
+is_PY3 = sys.version_info[0] >= 3
 if is_PY3:
     from functools import cmp_to_key
 
@@ -66,20 +67,19 @@ def to_utf8(text):
     if not text or is_PY3:
         return text
 
-    try:            # unicode or pure ascii
+    try:           # unicode or pure ascii
         return text.encode("utf8")
     except UnicodeDecodeError:
-        try:        # successful UTF-8 decode means it's pretty sure UTF-8
+        try:       # successful UTF-8 decode means it's pretty sure UTF-8
             text.decode("utf8")
             return text
         except UnicodeDecodeError:
-            try:    # get desperate; and yes,
-                    # this has a western hemisphere bias
+            try:   # get desperate; and yes, this has a western hemisphere bias
                 return text.decode("cp1252").encode("utf8")
             except UnicodeDecodeError:
                 pass
 
-    return text     # return unchanged, hope for the best
+    return text    # return unchanged, hope for the best
 
 
 class Config(object):
@@ -90,7 +90,7 @@ class Config(object):
     # Defaults
     node = 'latest'
     npm = 'latest'
-    with_npm = True if is_WIN or is_CYGWIN else False
+    with_npm = False
     jobs = '2'
     without_ssl = False
     debug = False
@@ -230,6 +230,11 @@ def parse_args(check=True):
         action='store_true', dest='io', default=False,
         help='Use iojs instead of nodejs.')
 
+    parser.add_option(
+        '--mirror',
+        action="store", dest='mirror',
+        help='Set mirror server of nodejs.org or iojs.org to download from.')
+
     if not is_WIN:
         parser.add_option(
             '-j', '--jobs', dest='jobs', default=Config.jobs,
@@ -380,13 +385,17 @@ def mkdir(path):
         logger.debug(' * Directory %s already exists', path)
 
 
+def make_executable(filename):
+    mode_0755 = (stat.S_IRWXU | stat.S_IXGRP |
+                 stat.S_IRGRP | stat.S_IROTH | stat.S_IXOTH)
+    os.chmod(filename, mode_0755)
+
+
 # noinspection PyArgumentList
 def writefile(dest, content, overwrite=True, append=False):
     """
     Create file and write content in it
     """
-    mode_0755 = (stat.S_IRWXU | stat.S_IXGRP |
-                 stat.S_IRGRP | stat.S_IROTH | stat.S_IXOTH)
     content = to_utf8(content)
     if is_PY3 and type(content) != bytes:
         content = bytes(content, 'utf-8')
@@ -394,7 +403,7 @@ def writefile(dest, content, overwrite=True, append=False):
         logger.debug(' * Writing %s ... ', dest, extra=dict(continued=True))
         with open(dest, 'wb') as f:
             f.write(content)
-        os.chmod(dest, mode_0755)
+        make_executable(dest)
         logger.debug('done.')
         return
     else:
@@ -512,7 +521,8 @@ def get_node_bin_url(version):
         'arch': archmap[platform.machine()],
     }
     if is_WIN or is_CYGWIN:
-        filename = 'win-%(arch)s/node.exe' % sysinfo
+        postfix = '-win-%(arch)s.zip' % sysinfo
+        filename = '%s-v%s%s' % (get_binary_prefix(), version, postfix)
     else:
         postfix = '-%(system)s-%(arch)s.tar.gz' % sysinfo
         filename = '%s-v%s%s' % (get_binary_prefix(), version, postfix)
@@ -543,18 +553,24 @@ def download_node_src(node_url, src_dir, opt, prefix):
     logger.info('.', extra=dict(continued=True))
 
     if is_WIN or is_CYGWIN:
-        writefile(join(src_dir, 'node.exe'), dl_contents.read())
+        ctx = zipfile.ZipFile(dl_contents)
+        members = operator.methodcaller('namelist')
+        member_name = lambda s: s  # noqa: E731
     else:
-        with tarfile_open(fileobj=dl_contents) as tarfile_obj:
-            member_list = tarfile_obj.getmembers()
-            extract_list = []
-            for member in member_list:
-                node_ver = opt.node.replace('.', '\.')
-                rexp_string = "%s-v%s[^/]*/(README\.md|CHANGELOG\.md|LICENSE)"\
-                    % (prefix, node_ver)
-                if re.match(rexp_string, member.name) is None:
-                    extract_list.append(member)
-            tarfile_obj.extractall(src_dir, extract_list)
+        ctx = tarfile_open(fileobj=dl_contents)
+        members = operator.methodcaller('getmembers')
+        member_name = operator.attrgetter('name')
+
+    with ctx as archive:
+        node_ver = re.escape(opt.node)
+        rexp_string = r"%s-v%s[^/]*/(README\.md|CHANGELOG\.md|LICENSE)"\
+            % (prefix, node_ver)
+        extract_list = [
+            member
+            for member in members(archive)
+            if re.match(rexp_string, member_name(member)) is None
+        ]
+        archive.extractall(src_dir, extract_list)
 
 
 def urlopen(url):
@@ -592,20 +608,26 @@ def copy_node_from_prebuilt(env_dir, src_dir, node_version):
     logger.info('.', extra=dict(continued=True))
     prefix = get_binary_prefix()
     if is_WIN:
-        src_exe = join(src_dir, 'node.exe')
-        dst_exe = join(env_dir, 'Scripts', 'node.exe')
-        mkdir(join(env_dir, 'Scripts'))
-        callit(['copy', '/Y', '/L', src_exe, dst_exe], False, True)
+        dest = join(env_dir, 'Scripts')
+        os.makedirs(dest)
     elif is_CYGWIN:
-        mkdir(join(env_dir, 'bin'))
+        dest = join(env_dir, 'bin')
+        os.makedirs(dest)
+        # write here to avoid https://bugs.python.org/issue35650
         writefile(join(env_dir, 'bin', 'node'), CYGWIN_NODE)
-        src_exe = join(src_dir, 'node.exe')
-        dst_exe = join(env_dir, 'bin', 'node.exe')
-        callit(['cp', '-a', src_exe, dst_exe], True, env_dir)
     else:
-        src_folder_tpl = src_dir + to_utf8('/%s-v%s*' % (prefix, node_version))
-        for src_folder in glob.glob(src_folder_tpl):
-            copytree(src_folder, env_dir, True)
+        dest = env_dir
+
+    src_folder_tpl = src_dir + to_utf8('/%s-v%s*' % (prefix, node_version))
+    src_folder, = glob.glob(src_folder_tpl)
+    copytree(src_folder, dest, True)
+
+    if is_CYGWIN:
+        for filename in ('npm', 'npx', 'node.exe'):
+            filename = join(env_dir, 'bin', filename)
+            if os.path.exists(filename):
+                make_executable(filename)
+
     logger.info('.', extra=dict(continued=True))
 
 
@@ -666,7 +688,7 @@ def build_node_from_src(env_dir, src_dir, node_src_dir, opt):
 
 
 def get_binary_prefix():
-    return to_utf8('node' if src_domain == 'nodejs.org' else 'iojs')
+    return to_utf8('node' if not iojs_taken else 'iojs')
 
 
 def install_node(env_dir, src_dir, opt):
@@ -674,6 +696,15 @@ def install_node(env_dir, src_dir, opt):
     Download source code for node.js, unpack it
     and install it in virtual environment.
     """
+    try:
+        install_node_wrapped(env_dir, src_dir, opt)
+    except BaseException:
+        # this restores the newline suppressed by continued=True
+        logger.info('')
+        raise
+
+
+def install_node_wrapped(env_dir, src_dir, opt):
     env_dir = abspath(env_dir)
     prefix = get_binary_prefix()
     node_src_dir = join(src_dir, to_utf8('%s-v%s' % (prefix, opt.node)))
@@ -739,7 +770,7 @@ def install_npm_win(env_dir, src_dir, opt):
     """
     logger.info(' * Install npm.js (%s) ... ' % opt.npm,
                 extra=dict(continued=True))
-    npm_url = 'https://github.com/npm/cli/archive/%s.zip' % opt.npm
+    npm_url = 'https://github.com/npm/cli/archive/v%s.zip' % opt.npm
     npm_contents = io.BytesIO(urlopen(npm_url).read())
 
     bin_path = join(env_dir, 'Scripts')
@@ -769,7 +800,7 @@ def install_npm_win(env_dir, src_dir, opt):
                     join(env_dir, 'bin', 'npm-cli.js'))
         shutil.copytree(join(bin_path, 'node_modules'),
                         join(env_dir, 'bin', 'node_modules'))
-        npm_gh_url = 'https://raw.githubusercontent.com/npm/npm'
+        npm_gh_url = 'https://raw.githubusercontent.com/npm/cli'
         npm_bin_url = '{}/{}/bin/npm'.format(npm_gh_url, opt.npm)
         writefile(join(env_dir, 'bin', 'npm'), urlopen(npm_bin_url).read())
 
@@ -934,7 +965,7 @@ class GetsAHrefs(HTMLParser):
             self.hrefs.append(dict(attrs).get('href', ''))
 
 
-VERSION_RE = re.compile('\d+\.\d+\.\d+')
+VERSION_RE = re.compile(r'\d+\.\d+\.\d+')
 
 
 def _py2_cmp(a, b):
@@ -1059,10 +1090,16 @@ def main():
         logger.error('Installing system node.js on win32 is not supported!')
         exit(1)
 
+    global iojs_taken
+    global src_domain
+
     if opt.io:
-        global src_domain
+        iojs_taken = True
         src_domain = "iojs.org"
 
+    if opt.mirror:
+        src_domain = opt.mirror
+
     if not opt.node or opt.node.lower() == "latest":
         opt.node = get_last_stable_node_version()
 
@@ -1110,7 +1147,7 @@ export npm_config_prefix=__NODE_VIRTUAL_ENV__
 exec __SHIM_NODE__ "$@"
 """
 
-ACTIVATE_BAT = """\
+ACTIVATE_BAT = r"""
 @echo off
 set "NODE_VIRTUAL_ENV=__NODE_VIRTUAL_ENV__"
 if not defined PROMPT (
@@ -1156,7 +1193,7 @@ set NODE_VIRTUAL_ENV=
 :END
 """
 
-ACTIVATE_PS1 = """\
+ACTIVATE_PS1 = r"""
 function global:deactivate ([switch]$NonDestructive) {
     # Revert to original values
     if (Test-Path function:_OLD_VIRTUAL_PROMPT) {
@@ -1203,7 +1240,7 @@ copy-item env:PATH env:_OLD_VIRTUAL_PATH
 $env:PATH = "$env:NODE_VIRTUAL_ENV\Scripts;$env:PATH"
 """
 
-ACTIVATE_SH = """
+ACTIVATE_SH = r"""
 
 # This file must be used with "source bin/activate" *from bash*
 # you cannot run it directly
@@ -1384,7 +1421,7 @@ end
 
 function freeze -d 'Show a list of installed packages - like `pip freeze`'
     set -l NPM_VER (npm -v | cut -d '.' -f 1)
-    set -l RE "[a-zA-Z0-9\.\-]+@[0-9]+\.[0-9]+\.[0-9]+([\+\-][a-zA-Z0-9\.\-]+)*"
+    set -l RE "[a-zA-Z0-9\\.\\-]+@[0-9]+\\.[0-9]+\\.[0-9]+([\\+\\-][a-zA-Z0-9\\.\\-]+)*"
 
     if test "$NPM_VER" = "0"
         set -g NPM_LIST (npm list installed active >/dev/null ^/dev/null | \
@@ -1395,7 +1432,7 @@ function freeze -d 'Show a list of installed packages - like `pip freeze`'
             set NPM_LS "npm ls"
             set -e argv[1]
         end
-        set -l NPM_LIST (eval $NPM_LS | grep -E '^.{4}\w{1}' | \
+        set -l NPM_LIST (eval $NPM_LS | grep -E '^.{4}\\w{1}' | \
                                         grep -o -E "$re" | \
                                         grep -v npm)
     end
@@ -1458,7 +1495,7 @@ if test -z "$NODE_VIRTUAL_ENV_DISABLE_PROMPT"
 
     set -gx _OLD_NODE_FISH_PROMPT_OVERRIDE "$NODE_VIRTUAL_ENV"
 end
-"""  # noqa
+"""  # noqa: E501
 
 PREDEACTIVATE_SH = """
 if type -p deactivate_node > /dev/null; then deactivate_node;fi
